from typing import List, Optional, Dict, Any, Union, Callable

from dateutil.parser import parse as from_iso

from aurora_connector.helpers.Exceptions import AuroraParameterException, AuroraDatabaseException


DATATYPE_MAPPING = {
    "VARCHAR": str,
    "CHAR": str,
    "BINARY": bytes,
    "VARBINARY": bytes,
    "TINYBLOB": bytes,
    "TINYTEXT": str,
    "TEXT": str,
    "BLOB": bytes,
    "MEDIUMTEXT": str,
    "MEDIUMBLOB": str,
    "LONGTEXT": str,
    "LONGBLOB": str,
    "BIT": bytes,
    "TINYINT": int,
    "BOOL": bool,
    "BOOLEAN": bool,
    "SMALLINT": int,
    "MEDIUMINT": int,
    "INT": int,
    "INTEGER": int,
    "BIGINT": int,
    "FLOAT": float,
    "DOUBLE": float,
    "DECIMAL": float,
    "DEC": float,
    "DATETIME": from_iso
}


def format_field(field: Dict[str, Any]) -> Any:
    """
    Returns the value of a field in a record

    :param field: (dict)
    :return: (Any)
    """

    return list(field.values())[0]


def format_record(record: List[Dict[str, Any]], column_types: List[Callable]) -> List[Any]:
    """
    Returns the formatted record with correct typing.

    :param record: A list with the following format:
        record = [
            {
                value_name: value
            },
            .
            .
            .
        ]
        where value_name = "stringValue", "blobValue", etc
        (See format_value for more information)
    :param column_types: A list of datatype constructors where the constructor stored at position i
                         relates to the datatype in position i stored in the record. (list)
    :return: (list)
    """

    return [column_type(format_field(field)) for field, column_type in zip(record, column_types)]


def get_column_type(column_type: str) -> Callable:
    """
    Returns the related data type constructor for the MySQL database with name :param column_type:

    :param column_type: Column type name. e.g. VARCHAR, etc (string)
    :return: (callable)
    """

    if column_type not in DATATYPE_MAPPING:
        raise AuroraDatabaseException({
            "message": "Unsupported column type in result set.",
            "column_type": column_type
        })

    return DATATYPE_MAPPING.get(column_type)


def get_column_types(column_metadata: List[Dict]) -> List[Callable]:
    """
    Parses column_metadata to extract the column types.
    Returns a list of callables, where each callable is the data type constructor for the related column type.
    This allows us to parse column types such as DATETIME which aren't supported by the data api

    :param column_metadata: The columnMetadata returned from the data API (list)
    :return: (list)
    """

    return [
        get_column_type(column.get("typeName")) for column in column_metadata
    ]


def format_response(response: Dict) -> List[List]:
    """

    :param response: Response dictionary from data api for a query.
    :return: (list)
    """

    column_types = get_column_types(response.get("columnMetadata"))
    records = response.get("records")

    return [format_record(record, column_types) for record in records]


def format_value(value: Any) -> Dict[str, Any]:
    """
    Returns value of an sql parameter according to the data api requirements.

    :param value: value of parameter (Any)
    :return: (dict)
    """

    if isinstance(value, bytes):
        return {"blobValue": value}

    if isinstance(value, bool):
        return {"booleanValue": value}

    if isinstance(value, float):
        return {"doubleValue": value}

    if value is None:
        return {"isNull": True}

    if isinstance(value, int):
        return {"longValue": value}

    if isinstance(value, str):
        return {"stringValue": value}

    raise AuroraParameterException({
        "message": "Unsupported parameter datatype.",
        "parameter_value": value,
        "supported_types": [
            bytes, bool, float, int, str
        ]
    })


def format_sql_parameters(sql_parameters: Dict[str, Any]) -> List[Dict[str, Union[str, Dict[str, Any]]]]:
    """
    Formats the sql parameters according to the data api requirements.

    :param sql_parameters: A dictionary with the following format:
        sql_parameters = {
            parameter_name: value,
            .
            .
            .
        }
    :return: (list)
    """

    return [
        {"name": name, "value": format_value(value)} for name, value in sql_parameters.items()
    ]


def fetch_one(result_set: List[List]) -> Optional[List]:
    """
    Returns the first record returned from a query.
    If the result set is empty then None is returned

    :param result_set: a list of records returned from AuroraDatabase.query (list)
    :return: (list|None)
    """

    return result_set[0] if result_set else None



import setuptools

with open("README.md", "r") as file:
    long_description = file.read()

requires = [
    "boto3 >= 1.9.215",
    "python-dateutil"
]

packages = [
    "aurora_connector",
    "aurora_connector.helpers"
]

setuptools.setup(
    name="aurora-connector",
    version="0.0.1",
    author="Alistair O'Brien",
    author_email="alistair@duneroot.co.uk",
    description="A Python AWS Aurora Serverless Connector.",
    long_description=long_description,
    include_package_data=True,
    long_description_content_type="text/markdown",
    url="https://github.com/johnyob/Aurora-Connector",
    packages=packages,
    install_requires=requires,
)



from typing import Optional, Any, Dict, List

from aurora_connector.helpers.Utils import format_response, format_sql_parameters
from aurora_connector.helpers.Exceptions import AuroraDatabaseException

import boto3
RDS_CLIENT = boto3.client("rds-data")


class AuroraDatabase:

    def __init__(self, database_name: str, db_cluster_arn: str, db_credentials_secrets_store_arn: str) -> None:
        """
        Creates a programmatic interface for an AWS Aurora Serverless cluster

        :param database_name: The name of the database. Relates to AWS::RDS::Cluster property DatabaseName. (string)
        :param db_cluster_arn: The RDS cluster amazon resource name. (string)
        :param db_credentials_secrets_store_arn: The amazon resource name for the AWS::SecretsManager::Secret credentials store. (string)
        """

        self._database_name = database_name
        self._db_cluster_arn = db_cluster_arn
        self._db_credentials_secrets_store_arn = db_credentials_secrets_store_arn

        self._transaction_id = None

    @property
    def database_name(self) -> str:
        """
        Returns the name of the database.

        :return: (string)
        """
        return self._database_name

    @property
    def db_cluster_arn(self) -> str:
        """
        Returns the RDS cluster amazon resource name.

        :return: (string)
        """
        return self._db_cluster_arn

    @property
    def db_credentials_secrets_store_arn(self) -> str:
        """
        Returns the database secrets store amazon resource name.

        :return: (string)
        """
        return self._db_credentials_secrets_store_arn

    @property
    def transaction_id(self) -> Optional[str]:
        """
        Returns the transaction_id. If the database is not "in" a transaction then None is returned, otherwise
        a string containing the UUID for the transaction_id is returned

        :return: (string|None)
        """
        return self._transaction_id

    def _in_transaction(self) -> bool:
        """
        Returns whether the database is currently in a transaction.
        We define the database being "in" a transaction if and only if
        a transaction has been started but not committed or rolled back

        :return: (boolean)
        """

        return self._transaction_id is not None

    def _begin_transaction(self) -> None:
        """
        Begins an AWS Aurora Serverless transaction.
        If a transaction is successfully started then the AuroraDatabase property :property transaction_id:
        is assigned the transactionId returned from the Data API.

        Note that we cannot begin a new transaction while currently "in" a transaction.
        An AuroraDatabaseException is thrown if this occurs.

        :return: (None)
        """

        if self._in_transaction():
            raise AuroraDatabaseException({
                "message": "Cannot start a new transaction while currently being in a transaction.",
                "transaction_id": self.transaction_id
            })

        try:
            response = RDS_CLIENT.begin_transaction(
                database=self.database_name,
                resourceArn=self.db_cluster_arn,
                secretArn=self.db_credentials_secrets_store_arn
            )
            self._transaction_id = response.get("transactionId")
        except Exception as error:
            raise AuroraDatabaseException(error) from error

    def commit(self) -> None:
        """
        Commits an AWS Aurora Serverless transaction.
        If the transaction is successfully committed then the AuroraDatabase property :property transaction_id:
        is assigned None.

        Note that we cannot commit a transaction if we're not currently "in" a transaction.
        An AuroraDatabaseException is thrown if this occurs.

        :return: (None)
        """

        if not self._in_transaction():
            raise AuroraDatabaseException({
                "message": "Cannot commit a transaction if a transaction hasn't been initiated.",
                "transaction_id": self.transaction_id
            })

        try:
            RDS_CLIENT.commit_transaction(
                resourceArn=self.db_cluster_arn,
                secretArn=self.db_credentials_secrets_store_arn,
                transactionId=self.transaction_id
            )

            # Reset transaction_id
            self._transaction_id = None
        except Exception as error:

            # Panic. If error occurs while committing transaction then reset transaction_id and then raise exception
            self._transaction_id = None
            raise AuroraDatabaseException(error) from error

    def rollback(self) -> None:
        """
        Rolls back an AWS Aurora Serverless transaction.
        If the transaction is successfully rolled back then the AuroraDatabase property :property transaction_id:
        is assigned None.

        Note that we cannot rollback a transaction if we're not currently "in" a transaction.
        An AuroraDatabaseException is thrown if this occurs.

        :return: (None)
        """

        if not self._in_transaction():
            raise AuroraDatabaseException({
                "message": "Cannot rollback a transaction if a transaction hasn't been initiated.",
                "transaction_id": self.transaction_id
            })

        try:
            RDS_CLIENT.rollback_transaction(
                resourceArn=self.db_cluster_arn,
                secretArn=self.db_credentials_secrets_store_arn,
                transactionId=self.transaction_id
            )


            # Reset transaction_id
            self._transaction_id = None
        except Exception as error:

            # Panic. If error occurs while committing transaction then reset transaction_id and then raise exception
            self._transaction_id = None
            raise AuroraDatabaseException(error) from error

    def _execute_statement(self, sql: str, sql_parameters: Dict[str, Any]) -> None:
        """
        Executes SQL statement.
        This method is only for statements that should use transaction in order to maintain ACID.
        If we're not "in" a transaction then begin a new transaction and then execute the statement.

        :param sql: the SQL statement to run (string)
        :param sql_parameters: A dictionary of parameters in the format: {parameter_name, parameter_value, } (dict)
        :return: (None)
        """

        if not self._in_transaction():
            self._begin_transaction()

        try:
            RDS_CLIENT.execute_statement(
                database=self.database_name,
                resourceArn=self.db_cluster_arn,
                secretArn=self.db_credentials_secrets_store_arn,
                transactionId=self.transaction_id,
                sql=sql,
                parameters=format_sql_parameters(sql_parameters)
            )
        except Exception as error:
            raise AuroraDatabaseException(error) from error

    def query(self, sql: str, sql_parameters: Dict[str, Any] = {}) -> List[List]:
        """
        Queries the AWS Aurora Serverless cluster and returns the result set with the following format:
            result_set = [
                [column_1, column_2, ..., column_n],
                .
                .
                .
            ]

        :param sql: the SQL statement to run (string)
        :param sql_parameters: A dictionary of parameters in the format: {parameter_name, parameter_value, } (dict)
        :return: (list)
        """

        try:
            response = RDS_CLIENT.execute_statement(
                includeResultMetadata=True,
                database=self._database_name,
                resourceArn=self._db_cluster_arn,
                secretArn=self._db_credentials_secrets_store_arn,
                sql=sql,
                parameters=format_sql_parameters(sql_parameters)
            )

            return format_response(response)
        except Exception as error:
            raise AuroraDatabaseException(error) from error

    def insert(self, sql: str, sql_parameters: Dict[str, Any] = {}) -> None:
        """
        A wrapper for AuroraDatabase.execute. To be used when executing an INSERT INTO statement.

        :param sql: the SQL statement to run (string)
        :param sql_parameters: A dictionary of parameters in the format: {parameter_name, parameter_value, } (dict)
        :return: (None)
        """

        self.execute(sql, sql_parameters)

    def delete(self, sql: str, sql_parameters: Dict[str, Any] = {}) -> None:
        """
        A wrapper for AuroraDatabase.execute. To be used when executing an DELETE FROM statement.

        :param sql: the SQL statement to run (string)
        :param sql_parameters: A dictionary of parameters in the format: {parameter_name, parameter_value, } (dict)
        :return: (None)
        """

        self.execute(sql, sql_parameters)

    def update(self, sql: str, sql_parameters: Dict[str, Any] = {}) -> None:
        """
        A wrapper for AuroraDatabase.execute. To be used when executing an UPDATE statement.

        :param sql: the SQL statement to run (string)
        :param sql_parameters: A dictionary of parameters in the format: {parameter_name, parameter_value, } (dict)
        :return: (None)
        """

        self.execute(sql, sql_parameters)

    def execute(self, sql: str, sql_parameters: Dict[str, Any] = {}) -> None:
        """
        A public wrapper for AuroraDatabase._execute_statement. To be used when executing any statement that required ACID.

        :param sql: the SQL statement to run (string)
        :param sql_parameters: A dictionary of parameters in the format: {parameter_name, parameter_value, } (dict)
        :return: (None)
        """

        try:
            self._execute_statement(sql, sql_parameters)
        except Exception as error:
            raise AuroraDatabaseException(error) from error

